MODULE PipeRead;

        (********************************************************)
        (*                                                      *)
        (*           Receiving messages via a named pipe        *)
        (*                                                      *)
        (*    Started:        15 March 2004                     *)
        (*    Last edited:    10 May 2004                       *)
        (*    Status:         OK                                *)
        (*                                                      *)
        (********************************************************)

IMPORT OS2, TextIO, STextIO, Strings;

FROM SYSTEM IMPORT
    (* type *)  CARD16,
    (* proc *)  ADR;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM VIOTitle IMPORT
    (* proc *)  SetOurTitle;

FROM IOChan IMPORT
    (* type *)  ChanId;

FROM CtrlC IMPORT
    (* proc *)  SetBreakHandler;

FROM LowLevel IMPORT
    (* proc *)  ALLOCATE64, EVAL;

FROM Storage IMPORT
    (* proc *)  DEALLOCATE;

(************************************************************************)

CONST
    Nul = CHR(0);
    BufferSize = 256;

VAR
    Shutdown: BOOLEAN;

(************************************************************************)
(*                       MISCELLANEOUS UTILITIES                        *)
(************************************************************************)

PROCEDURE ["C"] ControlCHandler(): BOOLEAN;

    (* Intercepts a Ctrl/C from the keyboard. *)

    BEGIN
        Shutdown := TRUE;
        RETURN TRUE;
    END ControlCHandler;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Writes N in decimal to standard output. *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        STextIO.WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)

PROCEDURE WriteError (rc: CARDINAL);

    VAR rc2, length: CARDINAL;
        message, filename: ARRAY [0..255] OF CHAR;

    (* Writes an error message belonging to the Dosxxx API. *)

    BEGIN
        filename := "OSO001.MSG";
        rc2 := OS2.DosGetMessage (NIL, 0, message, 256, rc,
                                           filename, length);
        IF rc2 = 0 THEN
            message[length] := Nul;
            STextIO.WriteString (message);
        ELSE
            STextIO.WriteString ("Message fault, error ");
            WriteCard (rc2);
            STextIO.WriteLn;
        END (*IF*);
    END WriteError;

(************************************************************************)

PROCEDURE SetScreenSize (NumberOfRows, NumberOfCols: CARDINAL);

    (* Sets the number of rows and columsns in the VIO window. *)

    VAR vioModeInfoPtr: POINTER TO OS2.VIOMODEINFO;
        oldrow, oldcol: CARD16;
        Fill: ARRAY [0..1] OF CHAR;
        change: BOOLEAN;

    BEGIN
        OS2.VioGetCurPos (oldrow, oldcol, 0);
        ALLOCATE64 (vioModeInfoPtr, SIZE(OS2.VIOMODEINFO));
        vioModeInfoPtr^.cb := SIZE (OS2.VIOMODEINFO);
        OS2.VioGetMode (vioModeInfoPtr^, 0);
        change := FALSE;
        IF vioModeInfoPtr^.row <> NumberOfRows THEN
            vioModeInfoPtr^.row := NumberOfRows;
            change := TRUE;
        END (*IF*);
        IF vioModeInfoPtr^.col <> NumberOfCols THEN
            vioModeInfoPtr^.col := NumberOfCols;
            change := TRUE;
        END (*IF*);
        IF change THEN
            OS2.VioSetMode (vioModeInfoPtr^, 0);
            IF oldrow >= vioModeInfoPtr^.row THEN
                Fill[0] := ' ';  Fill[1] := CHR(7);
                OS2.VioScrollUp (1, 0, vioModeInfoPtr^.row - 1,
                                 vioModeInfoPtr^.col - 1, 1, Fill, 0);
                oldrow := vioModeInfoPtr^.row - 1;
            END (*IF*);
            OS2.VioSetCurPos (oldrow, oldcol, 0);
        END (*IF*);
        DEALLOCATE (vioModeInfoPtr, SIZE(OS2.VIOMODEINFO));
    END SetScreenSize;

(************************************************************************)
(*                   PARSING THE PROGRAM ARGUMENTS                      *)
(************************************************************************)

PROCEDURE ProcessArguments (VAR (*OUT*) PipeName: ARRAY OF CHAR);

    (* Reads program arguments "pipename rows cols" (all optional),     *)
    (* sets the screen size, returns the pipe name.                     *)

    VAR Buffer: ARRAY [0..BufferSize-1] OF CHAR;
        pos: CARDINAL;

    (********************************************************************)

    PROCEDURE SkipBlanks;

        (* Updates pos to skip over blanks in Buffer. *)

        BEGIN
            WHILE Buffer[pos] = ' ' DO
                INC (pos);
            END (*WHILE*);
        END SkipBlanks;

    (********************************************************************)

    PROCEDURE GetNumber (VAR (*INOUT*) N: CARDINAL);

        (* Reads an unsigned number starting at Buffer[pos], updates    *)
        (* pos.  N is left unchanged if there is no number there.       *)

        TYPE CharSet = SET OF CHAR;
        CONST Digits = CharSet{'0'..'9'};

        BEGIN
            IF Buffer[pos] IN Digits THEN
                N := 0;
                WHILE Buffer[pos] IN Digits DO
                    N := 10*N + (ORD(Buffer[pos]) - ORD('0'));
                    INC (pos);
                END (*WHILE*);
            END (*IF*);
        END GetNumber;

    (********************************************************************)

    VAR NumberOfRows, NumberOfCols: CARDINAL;
        args: ChanId;  found: BOOLEAN;

    BEGIN                              (* Body of ProcessArguments *)
        args := ArgChan();
        IF IsArgPresent() THEN
            TextIO.ReadString (args, Buffer);
        ELSE
            (*Buffer := "WeaselSMTPLog 12 50";*)
            Buffer := "WeaselSMTPLog 25 80";
        END (*IF*);
        pos := 0;
        Strings.FindNext (' ', Buffer, 0, found, pos);
        Strings.Assign ("\PIPE\", PipeName);
        Strings.Append (Buffer, PipeName);
        IF found THEN
            PipeName[pos+6] := Nul;
        END (*IF*);

        (* Read the remainder of the program argument to get    *)
        (* screen height, width.                                *)

        NumberOfRows := 25;
        NumberOfCols := 80;
        SkipBlanks;
        GetNumber (NumberOfRows);
        SkipBlanks;
        GetNumber (NumberOfCols);

        (* Set the VIO screen size. *)

        SetScreenSize (NumberOfRows, NumberOfCols);

        Strings.Assign ("Reading from ", Buffer);
        Strings.Append (PipeName, Buffer);
        SetOurTitle (Buffer);

    END ProcessArguments;

(************************************************************************)
(*                          THE MAIN JOB                                *)
(************************************************************************)

PROCEDURE DoTheJob;

    CONST
        OpenFlags = OS2.OPEN_ACTION_FAIL_IF_NEW
                    + OS2.OPEN_ACTION_OPEN_IF_EXISTS;
        Mode = OS2.OPEN_FLAGS_FAIL_ON_ERROR + OS2.OPEN_FLAGS_SEQUENTIAL
                + OS2.OPEN_SHARE_DENYNONE
                + OS2.OPEN_FLAGS_NOINHERIT
                + OS2.OPEN_ACCESS_READONLY;

    VAR rc, Action, NumberRead, try: CARDINAL;
        hpipe: OS2.HPIPE;
        Waiting: BOOLEAN;
        PipeName, Buffer: ARRAY [0..BufferSize-1] OF CHAR;

    BEGIN
        (* Read the argument to get a pipe name. *)

        ProcessArguments (PipeName);

        STextIO.WriteString ("Getting data from pipe ");
        STextIO.WriteString (PipeName);
        STextIO.WriteLn;

        (* Wait for the input pipe to be open. *)

        try := 0;
        REPEAT
            rc := OS2.DosOpen (PipeName, hpipe, Action, 0, 0, OpenFlags, Mode, NIL);
            Waiting := (rc = 3) OR (rc = 231);
            IF Waiting THEN
                INC (try);
                IF try > 15 THEN
                    try := 0;
                    STextIO.WriteString ("Waiting for input pipe to be available");
                    STextIO.WriteLn;
                END (*IF*);
                OS2.DosSleep (200);
            END (*IF*);
        UNTIL Shutdown OR NOT Waiting;

        (* Transfer the data to the screen. *)

        IF rc = 0 THEN
            REPEAT
                IF OS2.DosRead (hpipe, ADR(Buffer),
                                    BufferSize, NumberRead) <> 0 THEN
                    NumberRead := 0;
                ELSIF NumberRead > 0 THEN
                    IF NumberRead < BufferSize THEN
                        Buffer[NumberRead] := Nul;
                    END (*IF*);
                    STextIO.WriteString (Buffer);
                END (*IF*);
            UNTIL NumberRead = 0;
            OS2.DosClose (hpipe);
            IF Shutdown THEN
                STextIO.WriteString ("*** PROGRAM TERMINATED FROM KEYBOARD ***");
            ELSE
                STextIO.WriteString ("*** END OF DATA FROM PIPE ***");
            END (*IF*);
            STextIO.WriteLn;
        ELSE
            WriteError (rc);
            STextIO.WriteString ("Failed to open input pipe");
            STextIO.WriteLn;
        END (*IF*);
        STextIO.WriteString ("Finished.");
        STextIO.WriteLn;
    END DoTheJob;

(************************************************************************)

BEGIN
    EVAL (SetBreakHandler(ControlCHandler));
    DoTheJob;
END PipeRead.

