/********************************************************/
/*                                                      */
/*           Receiving messages via a named pipe        */
/*                                                      */
/*    Author:         Peter Moylan, peter@ozebelg.org   */
/*    Started:        12 April 2007                     */
/*    Last edited:    14 April 2007                     */
/*    Status:         Working                           */
/*                                                      */
/*  Usage:                                              */
/*       piperead pipename screenrows screencols        */
/*                                                      */
/*     All arguments are optional.  If they are missing */
/*     we assume "piperead FtpServerTranslog 25 80"     */
/*                                                      */
/*  This is a translation into Rexx of the PipeRead     */
/*  program originally written in Modula-2.             */
/*                                                      */
/********************************************************/


CALL RxFuncAdd SysLoadFuncs, rexxutil, sysloadfuncs
CALL SysLoadFuncs
CALL RxFuncAdd 'rxuinit','rxu','rxuinit'
CALL rxuinit

SIGNAL ON ERROR NAME ErrorExit
SIGNAL ON HALT NAME ErrorExit
stdout = D2C(256*256*256, 4)

/* Get the PipeName and screen size from command arguments. */

PARSE ARG PipeName Rows Cols

IF PipeName = "" THEN PipeName = "FtpServerTranslog"
PipeName = "\PIPE\"||PipeName

IF Rows = "" THEN Rows = 25
IF Cols = "" THEN Cols = 80
MODE CO||Cols","Rows

SAY "Reading from "PipeName
ShutDown = 0

/* Wait for the input pipe to be open. */

try = 0
DO UNTIL Shutdown | \Waiting
    code = RxOpen('status.', PipeName, 'NO', 'FSHnR')
    Waiting = code \= 0
    IF Waiting THEN DO
        try = try + 1
        IF try > 15 THEN DO
            try = 0
            SAY "Waiting for input pipe to be available"
        END /*IF*/;
        CALL SysSleep 0.5
    END /*IF*/;
END /* DO UNTIL */

hpipe = status.1

/* Transfer the data to standard output (usually the screen). */

IF code = 0 THEN
    DO
        DO UNTIL NumberRead = 0
            code = RxRead('Buffer', hpipe, 256)
            NumberRead = LENGTH(Buffer)
            code = RxWrite(stdout, Buffer)
        END /* DO UNTIL */

        IF Shutdown THEN SAY "*** PROGRAM TERMINATED FROM KEYBOARD ***"
        ELSE SAY "*** END OF DATA FROM PIPE ***"
        CALL RxCloseH hpipe
    END
ELSE SAY "Failed to open input pipe"
SAY "Finished."

EXIT 0

ErrorExit:
    SIGNAL OFF ERROR
    IF (hpipe \= '') & (hpipe \= 'HPIPE') & (hpipe \= 'STATUS.1') THEN
        CALL RxCloseH hpipe
    EXIT 1

